from abc import abstractmethod
from datetime import UTC, date, datetime
from typing import (
    Any,
    ClassVar,
    Generic,
    Literal,
    TypedDict,
    TypeVar,
)

from pydantic import PrivateAttr, computed_field

from industrial_model.statements import Column

from .base import DBModelMetaclass, RootModel


class InstanceId(RootModel, metaclass=DBModelMetaclass):
    external_id: str
    space: str

    def __hash__(self) -> int:
        return hash((self.external_id, self.space))

    def __eq__(self, other: Any) -> bool:
        return (
            other is not None
            and isinstance(other, InstanceId)
            and self.external_id == other.external_id
            and self.space == other.space
        )

    def as_tuple(self) -> tuple[str, str]:
        return (self.space, self.external_id)


class EdgeContainer(InstanceId):
    type: InstanceId
    start_node: InstanceId
    end_node: InstanceId
    last_updated_time: int
    created_time: int

    @computed_field  # type: ignore[prop-decorator]
    @property
    def created_time_as_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.created_time / 1000, tz=UTC)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def last_updated_time_as_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.last_updated_time / 1000, tz=UTC)


TInstanceId = TypeVar("TInstanceId", bound=InstanceId)


class ViewInstanceConfig(TypedDict, total=False):
    view_external_id: str | None
    instance_spaces: list[str] | None
    instance_spaces_prefix: str | None
    view_code: str | None


class ViewInstance(InstanceId):
    view_config: ClassVar[ViewInstanceConfig] = ViewInstanceConfig()

    _edges: dict[str, list[EdgeContainer]] = PrivateAttr(default_factory=dict)

    @classmethod
    def get_view_external_id(cls) -> str:
        return cls.view_config.get("view_external_id") or cls.__name__

    def get_edge_metadata(self, property: Column | str | Any) -> list[EdgeContainer]:
        assert isinstance(property, Column | str), (
            f"Expected property to be Column, or str, got {type(property).__name__}"
        )
        identifier = property.property if isinstance(property, Column) else property

        edge_map_key = self.get_field_alias(identifier) or self.get_field_name(
            identifier
        )
        if not edge_map_key:
            return []
        return self._edges.get(edge_map_key, [])

    def generate_model_id(
        self,
        fields: list[str] | list[Column] | list[Any],
        view_code_as_prefix: bool = True,
        separator: str = "-",
    ) -> str:
        if not fields:
            raise ValueError("Fields list cannot be empty")
        field_values = self._get_field_values(fields)

        view_code = self.view_config.get("view_code")

        result = separator.join(field_values)
        return (
            f"{view_code}{separator}{result}"
            if view_code_as_prefix and view_code
            else result
        )

    def _get_field_values(
        self, fields: list[str] | list[Column] | list[Any]
    ) -> list[str]:
        field_values: list[str] = []
        for field in fields:
            if not isinstance(field, str | Column):
                field_type = type(field).__name__
                raise TypeError(
                    f"Expected field to be a string or Column, got {field_type}"
                )
            field_name = self.get_field_name(
                field.property if isinstance(field, Column) else field
            )
            if field_name is None:
                raise ValueError(f"Field {field} not found in the model")
            field_entry = getattr(self, field_name)

            field_entry_str = ""
            if isinstance(field_entry, str):
                field_entry_str = field_entry
            elif isinstance(field_entry, date | datetime):
                field_entry_str = field_entry.isoformat()
            elif isinstance(field_entry, InstanceId):
                field_entry_str = field_entry.external_id
            else:
                field_entry_str = str(field_entry)
            field_values.append(field_entry_str.replace(" ", ""))
        return field_values


class WritableViewInstance(ViewInstance):
    @abstractmethod
    def edge_id_factory(
        self, target_node: TInstanceId, edge_type: InstanceId
    ) -> InstanceId:
        raise NotImplementedError(
            "edge_id_factory method must be implemented in subclasses"
        )


class AggregatedViewInstance(RootModel, metaclass=DBModelMetaclass):
    view_config: ClassVar[ViewInstanceConfig] = ViewInstanceConfig()

    value: float

    @classmethod
    def get_view_external_id(cls) -> str:
        return cls.view_config.get("view_external_id") or cls.__name__

    @classmethod
    def get_group_by_fields(cls) -> list[str]:
        group_by_fields: set[str] = set()
        for key, field_info in cls.model_fields.items():
            if key == "value":
                continue
            group_by_fields.add(field_info.alias or key)

        return list(group_by_fields)


TViewInstance = TypeVar("TViewInstance", bound=ViewInstance)
TWritableViewInstance = TypeVar("TWritableViewInstance", bound=WritableViewInstance)
TAggregatedViewInstance = TypeVar(
    "TAggregatedViewInstance", bound=AggregatedViewInstance
)


class PaginatedResult(RootModel, Generic[TViewInstance]):
    data: list[TViewInstance]
    has_next_page: bool
    next_cursor: str | None

    def first_or_default(self) -> TViewInstance | None:
        return self.data[0] if self.data else None


ValidationMode = Literal["raiseOnError", "ignoreOnError"]



import datetime
from abc import ABC
from dataclasses import dataclass
from typing import Any

from industrial_model.constants import (
    BOOL_EXPRESSION_OPERATORS,
    LEAF_EXPRESSION_OPERATORS,
)

RANGE_SUPPORTED_VALUES = str | int | float
LIST_SUPPORTED_VALUES = list[str] | list[int] | list[float] | list[dict[str, str]]
SUPPORTED_VALUES = (
    RANGE_SUPPORTED_VALUES | bool | dict[str, str] | LIST_SUPPORTED_VALUES
)


class Expression(ABC):
    operator: str


@dataclass(frozen=True)
class BoolExpression(Expression):
    filters: list[Expression]
    operator: BOOL_EXPRESSION_OPERATORS

    def __and__(self, other: "BoolExpression | LeafExpression") -> bool:
        return and_(self, other)

    def __or__(self, other: "BoolExpression | LeafExpression") -> bool:
        return or_(self, other)


@dataclass(frozen=True)
class LeafExpression(Expression):
    property: str
    operator: LEAF_EXPRESSION_OPERATORS
    value: Any

    def __and__(self, other: "BoolExpression | LeafExpression") -> bool:
        return and_(self, other)

    def __or__(self, other: "BoolExpression | LeafExpression") -> bool:
        return or_(self, other)


class Column:
    def __init__(self, property: Any):
        assert isinstance(property, str | Column)
        property_ = property.property if isinstance(property, Column) else property
        self.property: str = property_

    def __hash__(self) -> int:
        return hash(self.property)

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return self.not_exists_()

        return self.equals_(other)

    def __ne__(self, other: Any) -> bool:
        if other is None:
            return self.exists_()

        return self.not_("==", other)

    def __lt__(self, other: Any) -> bool:
        return self.lt_(other)

    def __le__(self, other: Any) -> bool:
        return self.lte_(other)

    def __gt__(self, other: Any) -> bool:
        return self.gt_(other)

    def __ge__(self, other: Any) -> bool:
        return self.gte_(other)

    def equals_(self, other: SUPPORTED_VALUES) -> bool:
        return self._compare("==", other)

    def prefix(self, other: str) -> bool:
        return self._compare("prefix", other)

    def lt_(self, other: RANGE_SUPPORTED_VALUES | datetime.datetime) -> bool:
        return self._compare("<", other)

    def lte_(self, other: RANGE_SUPPORTED_VALUES | datetime.datetime) -> bool:
        return self._compare("<=", other)

    def gt_(self, other: RANGE_SUPPORTED_VALUES | datetime.datetime) -> bool:
        return self._compare(">", other)

    def gte_(self, other: RANGE_SUPPORTED_VALUES | datetime.datetime) -> bool:
        return self._compare(">=", other)

    def in_(self, other: LIST_SUPPORTED_VALUES) -> bool:
        return self._compare("in", other)

    def contains_all_(self, other: LIST_SUPPORTED_VALUES) -> bool:
        return self._compare("containsAll", other)

    def contains_any_(self, other: LIST_SUPPORTED_VALUES) -> bool:
        return self._compare("containsAny", other)

    def exists_(self) -> bool:
        return self._compare("exists", None)

    def not_exists_(self) -> bool:
        return self.not_("exists", None)

    def not_(self, operator: LEAF_EXPRESSION_OPERATORS, other: Any) -> bool:
        return BoolExpression(
            operator="not",
            filters=[self._compare(operator, other)],  # type: ignore
        )

    def nested_(self, expression: bool | LeafExpression | BoolExpression) -> bool:
        if not isinstance(expression, Expression):
            raise ValueError("Invalid expression")

        return self._compare("nested", expression)

    def _compare(self, operator: LEAF_EXPRESSION_OPERATORS, value: Any) -> bool:
        if isinstance(value, Column):
            raise ValueError("can not compare two columns in a graphdb")

        return LeafExpression(property=self.property, operator=operator, value=value)  # type: ignore


def _unwrap_expressions(
    *expressions: bool | LeafExpression | BoolExpression,
) -> list[Expression]:
    filters: list[Expression] = []
    for expression in expressions:
        assert isinstance(expression, Expression)
        filters.append(expression)
    return filters


def and_(*expressions: bool | LeafExpression | BoolExpression) -> bool:
    return BoolExpression(operator="and", filters=_unwrap_expressions(*expressions))  # type: ignore


def or_(*expressions: bool | LeafExpression | BoolExpression) -> bool:
    return BoolExpression(operator="or", filters=_unwrap_expressions(*expressions))  # type: ignore


def not_(*expressions: bool | LeafExpression | BoolExpression) -> bool:
    return BoolExpression(operator="not", filters=_unwrap_expressions(*expressions))  # type: ignore


def col(property: Any) -> Column:
    return Column(property)



from dataclasses import dataclass, field
from typing import Any, Generic, Literal, Self, TypeVar

from industrial_model.constants import DEFAULT_LIMIT, SORT_DIRECTION

from .expressions import (
    BoolExpression,
    Column,
    Expression,
    LeafExpression,
    and_,
    col,
    not_,
    or_,
)

T = TypeVar("T")
AggregateTypes = Literal["count", "avg", "min", "max", "sum"]
SearchOperationTypes = Literal["OR", "AND"]


def _create_column(property: str | Column | Any) -> Column:
    return property if isinstance(property, Column) else Column(property)


@dataclass
class BaseStatementValues:
    where_clauses: list[Expression] = field(init=False, default_factory=list)
    where_edge_clauses: list[tuple[Column, list[Expression]]] = field(
        init=False, default_factory=list
    )
    sort_clauses: list[tuple[Column, SORT_DIRECTION]] = field(
        init=False, default_factory=list
    )
    limit: int = field(init=False, default=DEFAULT_LIMIT)
    cursor: str | None = field(init=False, default=None)

    query: str | None = field(init=False, default=None)
    query_properties: list[str] | None = field(init=False, default=None)
    search_operator: SearchOperationTypes | None = field(init=False, default=None)


@dataclass
class BaseStatement(Generic[T]):
    entity: type[T] = field(init=True)
    _values: BaseStatementValues = field(
        init=False, default_factory=lambda: BaseStatementValues()
    )

    def where(self, *expressions: bool | Expression) -> Self:
        for expression in expressions:
            assert isinstance(expression, Expression)
            self._values.where_clauses.append(expression)
        return self

    def limit(self, limit: int) -> Self:
        self._values.limit = limit
        return self

    def asc(self, property: str | Column | Any) -> Self:
        return self.sort(property, "ascending")

    def desc(self, property: str | Column | Any) -> Self:
        return self.sort(property, "descending")

    def sort(self, property: str | Column | Any, direction: SORT_DIRECTION) -> Self:
        self._values.sort_clauses.append(
            (
                _create_column(property),
                direction,
            )
        )
        return self

    def get_values(self) -> BaseStatementValues:
        return self._values


@dataclass
class Statement(BaseStatement[T]):
    def cursor(self, cursor: str | None) -> Self:
        self._values.cursor = cursor
        return self

    def where_edge(
        self, property: str | Column | Any, *expressions: bool | Expression
    ) -> Self:
        if not expressions:
            return self

        expressions_: list[Expression] = []
        for expression in expressions:
            assert isinstance(expression, Expression)
            expressions_.append(expression)

        self._values.where_edge_clauses.append(
            (
                _create_column(property),
                expressions_,
            )
        )
        return self


@dataclass
class SearchStatement(BaseStatement[T]):
    def query_by(
        self,
        query: str,
        query_properties: list[Column] | list[str] | list[Any] | None = None,
        operation: SearchOperationTypes | None = None,
    ) -> Self:
        self._values.query = query
        self._values.query_properties = (
            [
                prop
                if isinstance(prop, str)
                else prop.property
                if isinstance(prop, Column)
                else str(prop)
                for prop in query_properties
            ]
            if query_properties
            else None
        )
        self._values.search_operator = operation
        return self


@dataclass
class AggregationStatementValues:
    group_by_properties: list[Column] | None = field(init=False, default=None)
    aggregation_property: Column = field(init=False, default=Column("externalId"))
    where_clauses: list[Expression] = field(init=False, default_factory=list)
    limit: int = field(init=False, default=-1)


@dataclass
class AggregationStatement(Generic[T]):
    entity: type[T] = field(init=True)
    aggregate: AggregateTypes = field(init=True)

    _values: AggregationStatementValues = field(
        init=False, default_factory=lambda: AggregationStatementValues()
    )

    def group_by(self, *property: str | Column | Any) -> Self:
        self._values.group_by_properties = [_create_column(prop) for prop in property]
        return self

    def aggregate_by(self, property: str | Column | Any) -> Self:
        self._values.aggregation_property = _create_column(property)
        return self

    def where(self, *expressions: bool | Expression) -> Self:
        for expression in expressions:
            assert isinstance(expression, Expression)
            self._values.where_clauses.append(expression)
        return self

    def limit(self, limit: int) -> Self:
        self._values.limit = limit
        return self

    def get_values(self) -> AggregationStatementValues:
        return self._values


def select(entity: type[T]) -> Statement[T]:
    return Statement(entity)


def aggregate(
    entity: type[T],
    aggregate: AggregateTypes | None = "count",
) -> AggregationStatement[T]:
    return AggregationStatement(entity=entity, aggregate=aggregate or "count")


def search(entity: type[T]) -> SearchStatement[T]:
    return SearchStatement(entity)


__all__ = [
    "aggregate",
    "AggregationStatement",
    "Statement",
    "select",
    "search",
    "SearchStatement",
    "Column",
    "col",
    "Expression",
    "LeafExpression",
    "BoolExpression",
    "and_",
    "not_",
    "or_",
]



from typing import Any

from .entities import EdgeContainer, TViewInstance, ViewInstance


def include_edges(
    root_item: dict[str, Any], validated_root_item: TViewInstance
) -> None:
    if not isinstance(root_item, dict) or not isinstance(
        validated_root_item, ViewInstance
    ):
        return

    for property_, value in root_item.items():
        field_key = validated_root_item.get_field_name(property_)
        if not field_key:
            continue
        _handle_edge_property(validated_root_item, value, field_key)
    _include_edges(root_item, validated_root_item)


def _handle_edge_property(
    validated_root_item: TViewInstance, value: Any, field_key: str
) -> None:
    attr_value = getattr(validated_root_item, field_key, None)

    if isinstance(value, dict):
        _handle_dict_property(value, attr_value)
        return

    if (
        not isinstance(value, list)
        or not isinstance(attr_value, list)
        or len(value) != len(attr_value)
    ):
        return

    for index, value_item in enumerate(value):
        entry = attr_value[index]
        _handle_dict_property(value_item, entry)


def _handle_dict_property(value: Any, attr_value: Any) -> None:
    if isinstance(value, dict) and isinstance(attr_value, ViewInstance):
        include_edges(value, attr_value)


def _include_edges(item: dict[str, Any], validated_item: TViewInstance) -> None:
    if "_edges" not in item or not isinstance(item["_edges"], dict):
        return
    entries: dict[str, list[EdgeContainer]] = {}
    for property_, edges in item["_edges"].items():
        if not edges or not isinstance(edges, list):
            continue

        assert isinstance(edges[0], EdgeContainer)
        entries[property_] = edges
    validated_item._edges = entries



from cognite.client.data_classes.data_modeling import InstanceSort, MappedProperty, View

from industrial_model.cognite_adapters.utils import get_property_ref
from industrial_model.constants import SORT_DIRECTION
from industrial_model.statements.expressions import Column


class SortMapper:
    def map(
        self,
        sort_clauses: list[tuple[Column, SORT_DIRECTION]],
        root_view: View,
    ) -> list[InstanceSort]:
        return [
            InstanceSort(
                property=get_property_ref(column.property, root_view),
                direction=direction,
                nulls_first=self._is_nulls_first(column, root_view, direction),
            )
            for column, direction in sort_clauses
        ]

    def _is_nulls_first(
        self, column: Column, root_view: View, direction: SORT_DIRECTION
    ) -> bool:
        view_property = root_view.properties.get(column.property)

        if isinstance(view_property, MappedProperty) and view_property.source:
            return direction == "ascending"

        return direction == "descending"



from typing import Any

from industrial_model.models import RootModel, TAggregatedViewInstance, TViewInstance
from industrial_model.statements import (
    AggregateTypes,
    AggregationStatement,
    SearchOperationTypes,
    SearchStatement,
    Statement,
    aggregate,
    search,
    select,
)

from .utils import extract_base_statement_params


class BaseQuery(RootModel):
    def to_statement(self, entity: type[TViewInstance]) -> Statement[TViewInstance]:
        statement = select(entity)
        _set_base_statement_params(self, statement)

        return statement


class BasePaginatedQuery(BaseQuery):
    limit: int = 1000
    cursor: str | None = None

    def to_statement(self, entity: type[TViewInstance]) -> Statement[TViewInstance]:
        statement = super().to_statement(entity)
        statement.limit(self.limit)
        statement.cursor(self.cursor)

        return statement


class BaseSearchQuery(RootModel):
    query: str | None = None
    query_properties: list[Any] | None = None
    query_operator: SearchOperationTypes | None = None
    limit: int = 1000

    def to_statement(
        self, entity: type[TViewInstance]
    ) -> SearchStatement[TViewInstance]:
        statement = search(entity)
        _set_base_statement_params(self, statement)
        if self.query:
            statement.query_by(
                self.query,
                self.query_properties,
                self.query_operator,
            )
        statement.limit(self.limit)

        return statement


class BaseAggregationQuery(RootModel):
    aggregate: AggregateTypes | None = None
    group_by_properties: list[Any] | None = None
    aggregation_property: str | None = None
    limit: int | None = None

    def to_statement(
        self, entity: type[TAggregatedViewInstance]
    ) -> AggregationStatement[TAggregatedViewInstance]:
        statement = aggregate(entity, self.aggregate)

        _set_base_statement_params(self, statement)

        if self.group_by_properties:
            statement.group_by(*self.group_by_properties)

        if self.aggregation_property:
            statement.aggregate_by(self.aggregation_property)

        if self.limit:
            statement.limit(self.limit)

        return statement


def _set_base_statement_params(
    entity: RootModel,
    statement: Statement[TViewInstance]
    | SearchStatement[TViewInstance]
    | AggregationStatement[TAggregatedViewInstance],
) -> None:
    filters_, sort_params = extract_base_statement_params(entity)
    statement.where(*filters_)

    if isinstance(statement, AggregationStatement):
        return

    for sort_value, direction in sort_params:
        statement.sort(sort_value, direction)



from pydantic import BaseModel, ConfigDict


class DataModelId(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    external_id: str
    space: str
    version: str

    def as_tuple(self) -> tuple[str, str, str]:
        return self.space, self.external_id, self.version



from threading import Lock

from cognite.client import CogniteClient

from industrial_model.models import TViewInstance
from industrial_model.statements import (
    BoolExpression,
    Expression,
    LeafExpression,
    Statement,
    col,
)

SPACE_PROPERTY = "space"


class QueryOptimizer:
    def __init__(self, cognite_client: CogniteClient):
        self._all_spaces: list[str] | None = None
        self._cognite_client = cognite_client
        self._lock = Lock()

    def optimize(self, statement: Statement[TViewInstance]) -> None:
        instance_spaces = statement.entity.view_config.get("instance_spaces")
        instance_spaces_prefix = statement.entity.view_config.get(
            "instance_spaces_prefix"
        )

        if not instance_spaces and not instance_spaces_prefix:
            return

        if self._has_space_filter(statement.get_values().where_clauses):
            return

        filter_spaces = (
            self._find_spaces(instance_spaces_prefix) if instance_spaces_prefix else []
        )
        if instance_spaces:
            filter_spaces.extend(instance_spaces)

        if filter_spaces:
            statement.where(col(SPACE_PROPERTY).in_(filter_spaces))

    def _has_space_filter(self, where_clauses: list[Expression]) -> bool:
        for where_clause in where_clauses:
            if isinstance(where_clause, BoolExpression) and self._has_space_filter(
                where_clause.filters
            ):
                return True
            elif (
                isinstance(where_clause, LeafExpression)
                and where_clause.property == SPACE_PROPERTY
            ):
                return True

        return False

    def _find_spaces(self, instance_spaces_prefix: str) -> list[str]:
        all_spaces = self._load_spaces()

        return [
            space for space in all_spaces if space.startswith(instance_spaces_prefix)
        ]

    def _load_spaces(self) -> list[str]:
        all_spaces = self._all_spaces
        if all_spaces:
            return all_spaces

        with self._lock:
            if self._all_spaces:
                return self._all_spaces

            all_spaces = self._cognite_client.data_modeling.spaces.list(
                limit=-1
            ).as_ids()

            self._all_spaces = all_spaces
            return all_spaces



import logging
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import Edge, Node
from cognite.client.data_classes.data_modeling.query import (
    Query as CogniteQuery,
)
from cognite.client.data_classes.data_modeling.query import (
    QueryResult as CogniteQueryResult,
)

from industrial_model.config import DataModelId
from industrial_model.models import (
    TAggregatedViewInstance,
    TViewInstance,
    TWritableViewInstance,
)
from industrial_model.statements import (
    AggregationStatement,
    SearchStatement,
    Statement,
)

from .aggregation_mapper import AggregationMapper
from .optimizer import QueryOptimizer
from .query_mapper import QueryMapper
from .query_result_mapper import (
    QueryResultMapper,
)
from .search_mapper import SearchMapper
from .upsert_mapper import UpsertMapper
from .utils import (
    append_nodes_and_edges,
    get_query_for_dependencies_pagination,
    map_nodes_and_edges,
)
from .view_mapper import ViewMapper


class CogniteAdapter:
    def __init__(self, cognite_client: CogniteClient, data_model_id: DataModelId):
        self._cognite_client = cognite_client

        view_mapper = ViewMapper(cognite_client, data_model_id)
        self._optmizer = QueryOptimizer(cognite_client)
        self._query_mapper = QueryMapper(view_mapper)
        self._result_mapper = QueryResultMapper(view_mapper)
        self._upsert_mapper = UpsertMapper(view_mapper)
        self._aggregation_mapper = AggregationMapper(view_mapper)
        self._search_mapper = SearchMapper(view_mapper)

    def search(self, statement: SearchStatement[TViewInstance]) -> list[dict[str, Any]]:
        search_query = self._search_mapper.map(statement)
        data = self._cognite_client.data_modeling.instances.search(
            view=search_query.view.as_id(),
            query=search_query.query,
            filter=search_query.filter,
            properties=search_query.query_properties,
            limit=search_query.limit,
            sort=search_query.sort,
            operator=search_query.operator or "OR",
        )

        return self._result_mapper.nodes_to_dict(data)

    def query(
        self, statement: Statement[TViewInstance], all_pages: bool
    ) -> tuple[list[dict[str, Any]], str | None]:
        self._optmizer.optimize(statement)
        cognite_query = self._query_mapper.map(statement)
        view_external_id = statement.entity.get_view_external_id()

        data: list[dict[str, Any]] = []
        while True:
            query_result = self._cognite_client.data_modeling.instances.query(
                cognite_query
            )

            dependencies_data = self._query_dependencies_pages(
                cognite_query, query_result, view_external_id
            )

            query_result_data = append_nodes_and_edges(
                map_nodes_and_edges(query_result, cognite_query),
                dependencies_data,
            )

            page_result = self._result_mapper.map_nodes(
                view_external_id,
                query_result_data,
            )
            next_cursor = query_result.cursors.get(view_external_id)
            data.extend(page_result)

            last_page = (
                len(page_result) < statement.get_values().limit or not next_cursor
            )
            next_cursor_ = None if last_page else next_cursor
            cognite_query.cursors = {view_external_id: next_cursor_}

            if not all_pages or last_page:
                return data, next_cursor_

    def aggregate(
        self, statement: AggregationStatement[TAggregatedViewInstance]
    ) -> list[dict[str, Any]]:
        query = self._aggregation_mapper.map(statement)

        result = self._cognite_client.data_modeling.instances.aggregate(
            view=query.view.as_id(),
            aggregates=query.metric_aggregation,
            filter=query.filters,
            group_by=query.group_by_columns,
            limit=query.limit,
        )
        data: list[dict[str, Any]] = []
        for item in result:
            if not item.aggregates or item.aggregates[0].value is None:
                continue

            entry = item.group if item.group else {}
            entry["value"] = item.aggregates[0].value
            data.append(entry)
        return data

    def upsert(
        self,
        entries: list[TWritableViewInstance],
        replace: bool = False,
        remove_unset: bool = False,
    ) -> None:
        logger = logging.getLogger(__name__)
        operation = self._upsert_mapper.map(entries, remove_unset)

        for node_chunk in operation.chunk_nodes():
            logger.debug(
                f"Upserting {len(node_chunk)} nodes (replace={replace}, "
                "remove_unset={remove_unset})"
            )
            self._cognite_client.data_modeling.instances.apply(
                nodes=node_chunk,
                replace=replace,
            )

        for edge_chunk in operation.chunk_edges():
            logger.debug(
                f"Upserting {len(edge_chunk)} edges (replace={replace},"
                "remove_unset={remove_unset})"
            )
            self._cognite_client.data_modeling.instances.apply(
                edges=edge_chunk,
                replace=replace,
            )

        for edges_to_remove_chunk in operation.chunk_edges_to_delete():
            logger.debug(f"Deleting {len(edges_to_remove_chunk)} edges")
            self._cognite_client.data_modeling.instances.delete(
                edges=[item.as_tuple() for item in edges_to_remove_chunk],
            )

    def delete(self, nodes: list[TViewInstance]) -> None:
        self._cognite_client.data_modeling.instances.delete(
            nodes=[item.as_tuple() for item in nodes],
        )

    def _query_dependencies_pages(
        self,
        cognite_query: CogniteQuery,
        query_result: CogniteQueryResult,
        view_external_id: str,
    ) -> dict[str, list[Node | Edge]] | None:
        new_query = get_query_for_dependencies_pagination(
            cognite_query, query_result, view_external_id
        )
        if not new_query:
            return None

        new_query_result = self._cognite_client.data_modeling.instances.query(new_query)

        result = map_nodes_and_edges(new_query_result, new_query)

        nested_results = self._query_dependencies_pages(
            new_query, new_query_result, view_external_id
        )
        return append_nodes_and_edges(result, nested_results)



from pathlib import Path

from cognite.client import CogniteClient

from industrial_model.config import DataModelId
from industrial_model.models import (
    PaginatedResult,
    TAggregatedViewInstance,
    TViewInstance,
    TWritableViewInstance,
    ValidationMode,
)
from industrial_model.statements import (
    AggregationStatement,
    SearchStatement,
    Statement,
)

from ._internal import generate_engine_params
from .engine import Engine


class AsyncEngine:
    def __init__(
        self,
        cognite_client: CogniteClient,
        data_model_id: DataModelId,
    ):
        self._engine = Engine(cognite_client, data_model_id)

    async def search_async(
        self,
        statement: SearchStatement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> list[TViewInstance]:
        return await self._engine.search_async(statement, validation_mode)

    async def query_async(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> PaginatedResult[TViewInstance]:
        return await self._engine.query_async(statement, validation_mode)

    async def query_all_pages_async(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> list[TViewInstance]:
        return await self._engine.query_all_pages_async(statement, validation_mode)

    async def aggregate_async(
        self, statement: AggregationStatement[TAggregatedViewInstance]
    ) -> list[TAggregatedViewInstance]:
        return await self._engine.aggregate_async(statement)

    async def upsert_async(
        self,
        entries: list[TWritableViewInstance],
        replace: bool = False,
        remove_unset: bool = False,
    ) -> None:
        return await self._engine.upsert_async(entries, replace, remove_unset)

    async def delete_async(self, nodes: list[TViewInstance]) -> None:
        return await self._engine.delete_async(nodes)

    @classmethod
    def from_config_file(cls, config_file: str | Path) -> "AsyncEngine":
        client, dm_id = generate_engine_params(config_file)
        return AsyncEngine(client, dm_id)



import datetime
from typing import Any

from cognite.client.data_classes.data_modeling import (
    DirectRelationReference,
    EdgeApply,
    EdgeConnection,
    MappedProperty,
    NodeApply,
    NodeOrEdgeData,
)

from industrial_model.cognite_adapters.models import UpsertOperation
from industrial_model.models import (
    EdgeContainer,
    InstanceId,
    TWritableViewInstance,
)
from industrial_model.utils import datetime_to_ms_iso_timestamp

from .view_mapper import ViewMapper


class UpsertMapper:
    def __init__(self, view_mapper: ViewMapper):
        self._view_mapper = view_mapper

    def map(
        self, instances: list[TWritableViewInstance], remove_unset: bool
    ) -> UpsertOperation:
        nodes: dict[tuple[str, str], NodeApply] = {}
        edges: dict[tuple[str, str], EdgeApply] = {}
        edges_to_delete: dict[tuple[str, str], EdgeContainer] = {}

        for instance in instances:
            entry_nodes, entry_edges, entry_edges_to_delete = self._map_instance(
                instance, remove_unset
            )

            nodes[instance.as_tuple()] = entry_nodes
            edges.update({(item.space, item.external_id): item for item in entry_edges})
            edges_to_delete.update(
                {(item.space, item.external_id): item for item in entry_edges_to_delete}
            )

        return UpsertOperation(
            nodes=list(nodes.values()),
            edges=list(edges.values()),
            edges_to_delete=list(edges_to_delete.values()),
        )

    def _map_instance(
        self, instance: TWritableViewInstance, remove_unset: bool
    ) -> tuple[NodeApply, list[EdgeApply], list[EdgeContainer]]:
        view = self._view_mapper.get_view(instance.get_view_external_id())

        edges: list[EdgeApply] = []
        edges_to_delete: list[EdgeContainer] = []
        properties: dict[str, Any] = {}
        for property_name, property in view.properties.items():
            property_key = instance.get_field_name(property_name)
            if not property_key:
                continue

            if remove_unset and property_key not in instance.model_fields_set:
                continue

            entry = instance.__getattribute__(property_key)

            if isinstance(property, MappedProperty):
                properties[property_name] = self._get_mapped_property_value(entry)
            elif isinstance(property, EdgeConnection) and isinstance(entry, list):
                possible_entries = self._map_edges(instance, property, entry)

                previous_edges = {
                    item.as_tuple(): item
                    for item in instance._edges.get(property_name, [])
                }

                new_entries = [
                    edge
                    for edge_id, edge in possible_entries.items()
                    if edge_id not in previous_edges
                ]
                edges_to_delete.extend(
                    [
                        previous_edges[edge_id]
                        for edge_id in previous_edges
                        if edge_id not in possible_entries
                    ]
                )

                edges.extend(new_entries)

        node = NodeApply(
            external_id=instance.external_id,
            space=instance.space,
            sources=[NodeOrEdgeData(source=view.as_id(), properties=properties)],
        )

        return node, edges, edges_to_delete

    def _get_mapped_property_value(self, entry: Any) -> Any:
        if isinstance(entry, list):
            return [self._get_mapped_property_value(item) for item in entry]

        if isinstance(entry, InstanceId):
            return DirectRelationReference(
                space=entry.space, external_id=entry.external_id
            )
        if isinstance(entry, datetime.datetime):
            return datetime_to_ms_iso_timestamp(entry)
        if isinstance(entry, datetime.date):
            return entry.strftime("%Y-%m-%d")
        return entry

    def _map_edges(
        self,
        instance: TWritableViewInstance,
        property: EdgeConnection,
        values: list[Any],
    ) -> dict[tuple[str, str], EdgeApply]:
        edge_type = InstanceId.model_validate(property.type)

        result: dict[tuple[str, str], EdgeApply] = {}
        for value in values:
            if not isinstance(value, InstanceId):
                raise ValueError(
                    f"""Invalid value for edge property {property.name}:
                        Received {type(value)} | Expected: InstanceId"""
                )

            start_node, end_node = (
                (instance, value)
                if property.direction == "outwards"
                else (value, instance)
            )

            edge_id = instance.edge_id_factory(value, edge_type)

            result[edge_id.as_tuple()] = EdgeApply(
                external_id=edge_id.external_id,
                space=edge_id.space,
                type=property.type,
                start_node=start_node.as_tuple(),
                end_node=end_node.as_tuple(),
            )

        return result



from datetime import date, datetime
from typing import Any

import cognite.client.data_classes.filters as cdf_filters
from cognite.client.data_classes.data_modeling import (
    EdgeConnection,
    MappedProperty,
    View,
)

from industrial_model.cognite_adapters.utils import get_property_ref
from industrial_model.models.entities import InstanceId
from industrial_model.statements import (
    BoolExpression,
    Column,
    Expression,
    LeafExpression,
)
from industrial_model.utils import datetime_to_ms_iso_timestamp

from .view_mapper import ViewMapper


class FilterMapper:
    def __init__(self, view_mapper: ViewMapper):
        self._view_mapper = view_mapper

    def map(
        self, expressions: list[Expression], root_view: View
    ) -> list[cdf_filters.Filter]:
        result: list[cdf_filters.Filter] = []
        for expression in expressions:
            if isinstance(expression, BoolExpression):
                result.append(self._to_cdf_filter_bool(expression, root_view))
            elif isinstance(expression, LeafExpression):
                result.append(self._to_cdf_filter_leaf(expression, root_view))
            else:
                cls_name = expression.__class__.__name__
                raise ValueError(f"Expression not implemented {cls_name}")
        return result

    def map_edges(
        self,
        edges_expressions: list[tuple[Column, list[Expression]]],
        root_view: View,
        nested_separator: str,
    ) -> dict[str, list[cdf_filters.Filter]]:
        result_dict: dict[str, list[cdf_filters.Filter]] = {}

        for column, expressions in edges_expressions:
            view_property = root_view.properties.get(column.property)
            if not isinstance(view_property, EdgeConnection):
                raise ValueError(f"Property {column.property} is not an edge")

            filters = self.map(
                expressions,
                self._view_mapper.get_view(view_property.source.external_id),
            )
            result_key = root_view.external_id + nested_separator + column.property
            result_dict.setdefault(result_key, []).extend(filters)

        return result_dict

    def _to_cdf_filter_bool(
        self, expression: BoolExpression, root_view: View
    ) -> cdf_filters.Filter:
        arguments = self.map(expression.filters, root_view)

        if expression.operator == "and":
            return cdf_filters.And(*arguments)
        elif expression.operator == "or":
            return cdf_filters.Or(*arguments)
        elif expression.operator == "not":
            return cdf_filters.Not(*arguments)

        raise NotImplementedError(f"Operator {self.operator} not implemented")

    def _to_cdf_filter_leaf(
        self,
        expression: LeafExpression,
        root_view: View,
    ) -> cdf_filters.Filter:
        property_ref = get_property_ref(expression.property, root_view)

        value_ = expression.value

        value_ = self._handle_type_value_convertion(value_)

        if expression.operator == "==":
            return cdf_filters.Equals(property_ref, value_)
        elif expression.operator == "in":
            return cdf_filters.In(property_ref, value_)
        elif expression.operator == ">":
            return cdf_filters.Range(property_ref, gt=value_)
        elif expression.operator == ">=":
            return cdf_filters.Range(property_ref, gte=value_)
        elif expression.operator == "<":
            return cdf_filters.Range(property_ref, lt=value_)
        elif expression.operator == "<=":
            return cdf_filters.Range(property_ref, lte=value_)
        elif expression.operator == "nested":
            target_view = self._get_nested_target_view(expression.property, root_view)

            assert isinstance(value_, Expression)

            return cdf_filters.Nested(
                property_ref,
                self.map([value_], target_view)[0],
            )
        elif expression.operator == "exists":
            return cdf_filters.Exists(property_ref)
        elif expression.operator == "prefix":
            return cdf_filters.Prefix(property_ref, value_)
        elif expression.operator == "containsAll":
            return cdf_filters.ContainsAll(property_ref, value_)
        elif expression.operator == "containsAny":
            return cdf_filters.ContainsAny(property_ref, value_)
        raise NotImplementedError(f"Operator {expression.operator} not implemented")

    def _get_nested_target_view(self, property: str, root_view: View) -> View:
        view_definiton = root_view.properties[property]
        assert isinstance(view_definiton, MappedProperty)
        assert view_definiton.source
        return self._view_mapper.get_view(view_definiton.source.external_id)

    def _handle_type_value_convertion(self, value_: Any) -> Any:
        if isinstance(value_, datetime):
            return datetime_to_ms_iso_timestamp(value_)
        elif isinstance(value_, date):
            return value_.strftime("%Y-%m-%d")
        elif isinstance(value_, InstanceId):
            return value_.model_dump(mode="json", by_alias=True)
        elif isinstance(value_, list):
            return [self._handle_type_value_convertion(v) for v in value_]
        return value_



from collections.abc import Callable
from datetime import datetime
from typing import (
    ParamSpec,
    TypeVar,
)

from anyio import to_thread

T_Retval = TypeVar("T_Retval")
P = ParamSpec("P")


def datetime_to_ms_iso_timestamp(dt: datetime) -> str:
    if not isinstance(dt, datetime):
        raise ValueError(f"Expected datetime object, got {type(dt)}")
    if dt.tzinfo is None:
        dt = dt.astimezone()
    return dt.isoformat(timespec="milliseconds")


async def run_async(
    func: Callable[..., T_Retval],
    *args: object,
    cancellable: bool = False,
) -> T_Retval:
    return await to_thread.run_sync(func, *args, cancellable=cancellable)



import os
from pathlib import Path
from string import Template

import yaml
from cognite.client import CogniteClient

from industrial_model.config import DataModelId


def generate_engine_params(
    config_file: str | Path,
) -> tuple[CogniteClient, DataModelId]:
    file_path = Path(config_file) if isinstance(config_file, str) else config_file

    if not file_path.exists():
        raise FileNotFoundError(f"Configuration file {file_path} does not exist")

    env_sub_template = Template(file_path.read_text())
    file_env_parsed = env_sub_template.substitute(dict(os.environ))

    engine_config = yaml.safe_load(file_env_parsed)
    assert isinstance(engine_config, dict), (
        "Configuration file must contain a dictionary"
    )
    assert "cognite" in engine_config, "Configuration must contain 'cognite' section"
    assert "data_model" in engine_config, (
        "Configuration must contain 'data_model' section"
    )

    client = CogniteClient.load(engine_config["cognite"])
    dm_id = DataModelId.model_validate(engine_config["data_model"])
    return client, dm_id



from typing import Any

from industrial_model.constants import SORT_DIRECTION
from industrial_model.models import RootModel
from industrial_model.statements import Expression

from .params import BoolQueryParam, NestedQueryParam, QueryParam, SortParam


def extract_base_statement_params(
    entity: RootModel,
) -> tuple[list[Expression], list[tuple[Any, SORT_DIRECTION]]]:
    filters_: list[Expression] = []
    sort_params: list[tuple[Any, SORT_DIRECTION]] = []

    for key, item in entity.__class__.model_fields.items():
        values = getattr(entity, key)
        if values is None:
            continue
        for metadata_item in item.metadata:
            if isinstance(metadata_item, SortParam):
                sort_params.append((values, metadata_item.direction))
            elif isinstance(
                metadata_item, QueryParam | NestedQueryParam | BoolQueryParam
            ):
                filter_ = _extract_expression(metadata_item, values)
                if filter_ is not None:
                    filters_.append(filter_)
    return filters_, sort_params


def _extract_expression(
    metadata_item: QueryParam | NestedQueryParam | BoolQueryParam, values: Any
) -> Expression | None:
    if isinstance(metadata_item, QueryParam | NestedQueryParam):
        return metadata_item.to_expression(values)

    if not isinstance(values, RootModel):
        return None

    filters_, _ = extract_base_statement_params(values)
    return metadata_item.to_expression(filters_)



from dataclasses import dataclass
from typing import Any

from cognite.client.data_classes.data_modeling import (
    EdgeApply,
    NodeApply,
)

from industrial_model.models.entities import EdgeContainer

_PAGE_SIZE = 1000


@dataclass
class UpsertOperation:
    nodes: list[NodeApply]
    edges: list[EdgeApply]
    edges_to_delete: list[EdgeContainer]

    def chunk_nodes(self) -> list[list[NodeApply]]:
        return self._chunk_list(self.nodes)

    def chunk_edges(self) -> list[list[EdgeApply]]:
        return self._chunk_list(self.edges)

    def chunk_edges_to_delete(self) -> list[list[EdgeContainer]]:
        return self._chunk_list(self.edges_to_delete)

    def _chunk_list(self, entries: list[Any]) -> list[list[Any]]:
        data: list[list[Any]] = []
        for i in range(0, len(entries), _PAGE_SIZE):
            start = i
            end = i + _PAGE_SIZE
            data.append(entries[start:end])
        return data



from .base import RootModel
from .entities import (
    AggregatedViewInstance,
    EdgeContainer,
    InstanceId,
    PaginatedResult,
    TAggregatedViewInstance,
    TViewInstance,
    TWritableViewInstance,
    ValidationMode,
    ViewInstance,
    ViewInstanceConfig,
    WritableViewInstance,
)
from .schemas import get_parent_and_children_nodes, get_schema_properties
from .utils import include_edges

__all__ = [
    "AggregatedViewInstance",
    "RootModel",
    "EdgeContainer",
    "InstanceId",
    "include_edges",
    "TAggregatedViewInstance",
    "TViewInstance",
    "TWritableViewInstance",
    "ViewInstance",
    "ValidationMode",
    "PaginatedResult",
    "ViewInstanceConfig",
    "get_schema_properties",
    "get_parent_and_children_nodes",
    "WritableViewInstance",
]



from collections import defaultdict
from enum import StrEnum
from typing import Any, TypedDict

from cognite.client.data_classes.data_modeling import (
    Edge,
    EdgeConnection,
    MappedProperty,
    Node,
    NodeList,
    View,
)
from cognite.client.data_classes.data_modeling.data_types import (
    ListablePropertyType,
)
from cognite.client.data_classes.data_modeling.instances import PropertyValue
from cognite.client.data_classes.data_modeling.views import (
    MultiReverseDirectRelation,
    SingleReverseDirectRelation,
)

from industrial_model.constants import EDGE_DIRECTION, EDGE_MARKER, NESTED_SEP
from industrial_model.models import EdgeContainer

from .view_mapper import ViewMapper


class ConnectionTypeEnum(StrEnum):
    DIRECT_RELATION = "DirectRelation"
    REVERSE_DIRECT_RELATION = "ReverseDirectRelation"
    EDGE = "Edge"


class _PropertyMapping(TypedDict):
    is_list: bool
    connection_type: ConnectionTypeEnum
    nodes: dict[tuple[str, str], list[Node]]
    edges: dict[tuple[str, str], list[Edge]]


class QueryResultMapper:
    def __init__(self, view_mapper: ViewMapper):
        self._view_mapper = view_mapper

    def map_nodes(
        self, root_node: str, query_result: dict[str, list[Node | Edge]]
    ) -> list[dict[str, Any]]:
        if root_node not in query_result:
            raise ValueError(f"{root_node} is not available in the query result")

        root_view = self._view_mapper.get_view(root_node)

        values = self._map_node_property(root_node, root_view, query_result)
        if not values:
            return []

        data = [node for nodes in values.values() for node in self.nodes_to_dict(nodes)]

        return data

    def nodes_to_dict(self, nodes: list[Node] | NodeList[Node]) -> list[dict[str, Any]]:
        return [self._node_to_dict(node) for node in nodes]

    def _map_node_property(
        self,
        key: str,
        view: View,
        query_result: dict[str, list[Node | Edge]],
        result_property_key: str | None = None,
    ) -> dict[tuple[str, str], list[Node]] | None:
        if key not in query_result:
            return None

        mappings = self._get_property_mappings(key, view, query_result)

        view_id = view.as_id()

        def get_node_id(node: Node) -> tuple[str, str]:
            if not result_property_key:
                return (node.space, node.external_id)

            entry = properties.get(result_property_key)
            if not isinstance(entry, dict):
                raise ValueError(f"Invalid result property key {result_property_key}")

            return entry.get("space", ""), entry.get("externalId", "")

        visited: set[tuple[str, str]] = set()
        result: defaultdict[tuple[str, str], list[Node]] = defaultdict(list)
        for node in query_result[key]:
            identify = (node.space, node.external_id)
            if not isinstance(node, Node) or identify in visited:
                continue

            visited.add(identify)
            properties = node.properties.get(view_id, {})
            if len(properties) == 0 and view_id not in node.properties:
                continue

            edges_mapping: dict[str, list[EdgeContainer]] = {}
            node_id = get_node_id(node)
            for mapping_key, mapping_value in mappings.items():
                element = properties.get(mapping_key)

                mapping_nodes = mapping_value.get("nodes", {})
                mapping_edges = mapping_value.get("edges", {})
                is_list = mapping_value.get("is_list", False)
                connection_type = mapping_value.get(
                    "connection_type",
                    ConnectionTypeEnum.DIRECT_RELATION,
                )

                if (
                    element is None
                    and connection_type == ConnectionTypeEnum.DIRECT_RELATION
                ):
                    continue

                element_keys = self._get_element_keys(node, element)

                node_entries = [
                    item
                    for element_key in element_keys
                    for item in mapping_nodes.get(element_key, [])
                ]
                if not node_entries:
                    if mapping_key in properties:
                        properties.pop(mapping_key)
                    continue

                entry_data = self.nodes_to_dict(node_entries)
                properties[mapping_key] = entry_data if is_list else entry_data[0]
                edge_entries = [
                    item
                    for element_key in element_keys
                    for item in mapping_edges.get(element_key, [])
                ]
                if edge_entries:
                    edges_mapping[mapping_key] = self._edges_to_model(edge_entries)
            properties["_edges"] = edges_mapping

            node.properties[view_id] = properties

            result[node_id].append(node)

        return dict(result)

    def _get_element_keys(
        self, node: Node, element: PropertyValue | None
    ) -> list[tuple[str, str]]:
        if isinstance(element, dict):
            return [(element.get("space", ""), element.get("externalId", ""))]

        if isinstance(element, list):
            return [
                (item.get("space", ""), item.get("externalId", ""))
                for item in element
                if isinstance(item, dict)
            ]

        return [(node.space, node.external_id)]

    def _get_property_mappings(
        self,
        key: str,
        view: View,
        query_result: dict[str, list[Node | Edge]],
    ) -> dict[str, _PropertyMapping]:
        mappings: dict[str, _PropertyMapping] = {}

        for property_name, property in view.properties.items():
            property_key = f"{key}{NESTED_SEP}{property_name}"

            nodes: dict[tuple[str, str], list[Node]] | None = None
            edges: dict[tuple[str, str], list[Edge]] | None = None
            is_list = False
            connection_type: ConnectionTypeEnum = ConnectionTypeEnum.DIRECT_RELATION

            if isinstance(property, MappedProperty) and property.source:
                nodes = self._map_node_property(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    query_result,
                )
                is_list = (
                    isinstance(property.type, ListablePropertyType)
                    and property.type.is_list
                )
                connection_type = ConnectionTypeEnum.DIRECT_RELATION
            elif isinstance(property, SingleReverseDirectRelation) and property.source:
                nodes = self._map_node_property(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    query_result,
                    property.through.property,
                )
                is_list = False
                connection_type = ConnectionTypeEnum.REVERSE_DIRECT_RELATION
            elif isinstance(property, MultiReverseDirectRelation) and property.source:
                nodes = self._map_node_property(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    query_result,
                    property.through.property,
                )
                is_list = True
                connection_type = ConnectionTypeEnum.REVERSE_DIRECT_RELATION

            elif isinstance(property, EdgeConnection) and property.source:
                nodes, edges = self._map_edge_property(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    query_result,
                    property.direction,
                )
                is_list = True
                connection_type = ConnectionTypeEnum.EDGE

            if nodes is not None:
                mappings[property_name] = _PropertyMapping(
                    is_list=is_list,
                    connection_type=connection_type,
                    nodes=nodes,
                    edges=edges or {},
                )

        return mappings

    def _map_edge_property(
        self,
        key: str,
        view: View,
        query_result: dict[str, list[Node | Edge]],
        edge_direction: EDGE_DIRECTION,
    ) -> tuple[
        dict[tuple[str, str], list[Node]] | None,
        dict[tuple[str, str], list[Edge]] | None,
    ]:
        edge_key = f"{key}{NESTED_SEP}{EDGE_MARKER}"
        if key not in query_result or edge_key not in query_result:
            return None, None

        nodes = self._map_node_property(key, view, query_result)
        if not nodes:
            return None, None

        visited: set[tuple[str, str]] = set()
        nodes_result: defaultdict[tuple[str, str], list[Node]] = defaultdict(list)
        edges_result: defaultdict[tuple[str, str], list[Edge]] = defaultdict(list)
        for edge in query_result[edge_key]:
            identify = (edge.space, edge.external_id)
            if not isinstance(edge, Edge) or identify in visited:
                continue

            visited.add(identify)
            entry_key, node_key = (
                (
                    edge.end_node.as_tuple(),
                    edge.start_node.as_tuple(),
                )
                if edge_direction == "inwards"
                else (edge.start_node.as_tuple(), edge.end_node.as_tuple())
            )
            edges_result[entry_key].append(edge)
            if node_item := nodes.get(node_key):
                nodes_result[entry_key].extend(node_item)

        return dict(nodes_result), dict(edges_result)

    def _edges_to_model(self, edges: list[Edge]) -> list[EdgeContainer]:
        return [EdgeContainer.model_validate(edge) for edge in edges]

    def _node_to_dict(self, node: Node) -> dict[str, Any]:
        entry = node.dump()
        properties: dict[str, dict[str, dict[str, Any]]] = entry.pop("properties") or {}
        for space_mapping in properties.values():
            for view_mapping in space_mapping.values():
                entry.update(view_mapping)

        return entry



from dataclasses import dataclass
from typing import Any

from industrial_model.constants import (
    BOOL_EXPRESSION_OPERATORS,
    LEAF_EXPRESSION_OPERATORS,
    SORT_DIRECTION,
)
from industrial_model.statements import BoolExpression, Expression, LeafExpression


@dataclass
class QueryParam:
    property: str
    operator: LEAF_EXPRESSION_OPERATORS

    def to_expression(self, value: Any) -> Expression:
        if self.operator == "nested":
            raise ValueError(
                "Nested operator not allowed in QueryParam - use NestedQueryParam"
            )

        if self.operator == "exists" and isinstance(value, bool) and not value:
            return BoolExpression(
                operator="not",
                filters=[
                    LeafExpression(
                        property=self.property, operator=self.operator, value=True
                    )
                ],
            )

        return LeafExpression(
            property=self.property,
            operator=self.operator,
            value=value,
        )


@dataclass
class NestedQueryParam:
    property: str
    value: QueryParam

    def to_expression(self, value: Any) -> Expression:
        return LeafExpression(
            property=self.property,
            operator="nested",
            value=self.value.to_expression(value),
        )


@dataclass
class BoolQueryParam:
    operator: BOOL_EXPRESSION_OPERATORS

    def to_expression(self, filters: list[Expression]) -> Expression:
        return BoolExpression(
            operator=self.operator,
            filters=filters,
        )


@dataclass
class SortParam:
    direction: SORT_DIRECTION



from threading import Lock

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import (
    View,
)

from industrial_model.config import DataModelId


class ViewMapper:
    def __init__(self, cognite_client: CogniteClient, data_model_id: DataModelId):
        self._cognite_client = cognite_client
        self._data_model_id = data_model_id
        self._views_as_dict: dict[str, View] | None = None

        self._lock = Lock()

    def get_view(self, view_external_id: str) -> View:
        views = self._load_views()
        if view_external_id not in views:
            raise ValueError(f"View {view_external_id} is not available in data model")

        return views[view_external_id]

    def _load_views(self) -> dict[str, View]:
        if self._views_as_dict:
            return self._views_as_dict

        with self._lock:
            if self._views_as_dict:
                return self._views_as_dict

            dm = self._cognite_client.data_modeling.data_models.retrieve(
                ids=self._data_model_id.as_tuple(),
                inline_views=True,
            ).latest_version()

            views = {view.external_id: view for view in dm.views}
            self._views_as_dict = views
            return views



from typing import Literal

from cognite.client.data_classes.data_modeling import (
    Edge,
    Node,
    NodeListWithCursor,
    View,
    ViewId,
)
from cognite.client.data_classes.data_modeling.query import (
    Query as CogniteQuery,
)
from cognite.client.data_classes.data_modeling.query import (
    QueryResult as CogniteQueryResult,
)
from cognite.client.data_classes.data_modeling.query import (
    ResultSetExpression,
)
from cognite.client.data_classes.data_modeling.query import (
    Select as CogniteSelect,
)

from industrial_model.constants import MAX_LIMIT
from industrial_model.models import (
    TViewInstance,
    get_parent_and_children_nodes,
)

NODE_PROPERTIES = {
    "externalId",
    "space",
    "createdTime",
    "deletedTime",
    "lastUpdatedTime",
}
INTANCE_TYPE = Literal["node", "edge"]


def get_property_ref(
    property: str, view: View | ViewId, instance_type: INTANCE_TYPE = "node"
) -> tuple[str, str, str] | tuple[str, str]:
    return (
        view.as_property_ref(property)
        if property not in NODE_PROPERTIES
        else (instance_type, property)
    )


def get_cognite_instance_ids(
    instance_ids: list[TViewInstance],
) -> list[dict[str, str]]:
    return [get_cognite_instance_id(instance_id) for instance_id in instance_ids]


def get_cognite_instance_id(instance_id: TViewInstance) -> dict[str, str]:
    return {"space": instance_id.space, "externalId": instance_id.external_id}


def get_query_for_dependencies_pagination(
    query: CogniteQuery,
    query_result: CogniteQueryResult,
    view_external_id: str,
) -> CogniteQuery | None:
    nodes_parent, nodes_children = get_parent_and_children_nodes(
        set(query_result.cursors.keys())
    )

    leaf_cursors = _get_leaf_cursors(
        query_result, view_external_id, nodes_parent, nodes_children
    )

    if not leaf_cursors:
        return None

    return _create_query(query, nodes_parent, nodes_children, leaf_cursors)


def map_nodes_and_edges(
    query_result: CogniteQueryResult, query: CogniteQuery
) -> dict[str, list[Node | Edge]]:
    result_schema = query.instance_type_by_result_expression()

    return {
        key: query_result.get_nodes(key).data
        if entity_type is NodeListWithCursor
        else query_result.get_edges(key).data
        for key, entity_type in result_schema.items()
        if entity_type is not None
    }


def append_nodes_and_edges(
    initial_dataset: dict[str, list[Node | Edge]],
    additional_dataset: dict[str, list[Node | Edge]] | None,
) -> dict[str, list[Node | Edge]]:
    if not additional_dataset:
        return initial_dataset
    for key, additional_data in additional_dataset.items():
        if key not in initial_dataset:
            initial_dataset[key] = []
        initial_dataset[key].extend(additional_data)
    return initial_dataset


def _create_query(
    previous_query: CogniteQuery,
    nodes_parent: dict[str, set[str]],
    nodes_children: dict[str, set[str]],
    leaf_cursors: dict[str, str],
) -> CogniteQuery:
    with_: dict[str, ResultSetExpression] = {}
    select_: dict[str, CogniteSelect] = {}
    final_cursors: dict[str, str | None] = {}

    for cursor_key, cursor_value in leaf_cursors.items():
        children = nodes_children.get(cursor_key, set())
        parent = nodes_parent.get(cursor_key, set())

        valid_keys = parent.union(children)
        valid_keys.add(cursor_key)

        with_.update({k: v for k, v in previous_query.with_.items() if k in valid_keys})
        select_.update(
            {k: v for k, v in previous_query.select.items() if k in valid_keys}
        )

        final_cursors.update(
            {k: v for k, v in previous_query.cursors.items() if k in parent}
        )
        final_cursors[cursor_key] = cursor_value

    return CogniteQuery(with_=with_, select=select_, cursors=final_cursors)


def _get_leaf_cursors(
    query_result: CogniteQueryResult,
    view_external_id: str,
    nodes_parent: dict[str, set[str]],
    nodes_children: dict[str, set[str]],
) -> dict[str, str]:
    target_cursors: dict[str, str] = {}
    for cursor_key, cursor_value in query_result.cursors.items():
        if (
            cursor_key == view_external_id
            or not cursor_value
            or len(query_result[cursor_key]) != MAX_LIMIT
        ):
            continue

        children = nodes_children.get(cursor_key, set())

        target_cursors_keys = target_cursors.keys()
        if len(children.intersection(target_cursors_keys)) > 0:
            continue

        parent = nodes_parent.get(cursor_key, set())
        cursor_to_remove = parent.intersection(target_cursors_keys)
        for cursor_key_to_remove in cursor_to_remove:
            target_cursors.pop(cursor_key_to_remove)

        target_cursors[cursor_key] = cursor_value
    return target_cursors



from .models import BaseAggregationQuery, BasePaginatedQuery, BaseQuery, BaseSearchQuery
from .params import BoolQueryParam, NestedQueryParam, QueryParam, SortParam

__all__ = [
    "BaseQuery",
    "BasePaginatedQuery",
    "BaseSearchQuery",
    "BaseAggregationQuery",
    "SortParam",
    "QueryParam",
    "NestedQueryParam",
    "BoolQueryParam",
]



from dataclasses import dataclass

import cognite.client.data_classes.filters as filters
from cognite.client.data_classes.data_modeling import InstanceSort, View

from industrial_model.models import TViewInstance
from industrial_model.statements import SearchOperationTypes, SearchStatement

from .filter_mapper import (
    FilterMapper,
)
from .sort_mapper import SortMapper
from .view_mapper import ViewMapper


@dataclass
class SearchQuery:
    view: View
    filter: filters.Filter | None
    query: str | None
    query_properties: list[str] | None
    limit: int
    sort: list[InstanceSort]
    operator: SearchOperationTypes | None


class SearchMapper:
    def __init__(self, view_mapper: ViewMapper):
        self._view_mapper = view_mapper
        self._filter_mapper = FilterMapper(view_mapper)
        self._sort_mapper = SortMapper()

    def map(self, statement: SearchStatement[TViewInstance]) -> SearchQuery:
        root_node = statement.entity.get_view_external_id()
        statement_values = statement.get_values()
        root_view = self._view_mapper.get_view(root_node)

        filters_ = self._filter_mapper.map(statement_values.where_clauses, root_view)

        sort_clauses = self._sort_mapper.map(statement_values.sort_clauses, root_view)
        for item in sort_clauses:
            item.nulls_first = None

        return SearchQuery(
            view=root_view,
            filter=filters.And(*filters_) if filters_ else None,
            query=statement_values.query,
            query_properties=statement_values.query_properties,
            limit=statement_values.limit,
            sort=sort_clauses,
            operator=statement_values.search_operator,
        )



from pathlib import Path
from typing import Any

from cognite.client import CogniteClient

from industrial_model.cognite_adapters import CogniteAdapter
from industrial_model.config import DataModelId
from industrial_model.models import (
    PaginatedResult,
    TAggregatedViewInstance,
    TViewInstance,
    TWritableViewInstance,
    ValidationMode,
    include_edges,
)
from industrial_model.statements import (
    AggregationStatement,
    SearchStatement,
    Statement,
)
from industrial_model.utils import run_async

from ._internal import generate_engine_params


class Engine:
    def __init__(
        self,
        cognite_client: CogniteClient,
        data_model_id: DataModelId,
    ):
        self._cognite_adapter = CogniteAdapter(cognite_client, data_model_id)

    def search(
        self,
        statement: SearchStatement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> list[TViewInstance]:
        data = self._cognite_adapter.search(statement)
        return self._validate_data(statement.entity, data, validation_mode)

    async def search_async(
        self,
        statement: SearchStatement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> list[TViewInstance]:
        return await run_async(self.search, statement, validation_mode)

    def query(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> PaginatedResult[TViewInstance]:
        data, next_cursor = self._cognite_adapter.query(statement, False)

        return PaginatedResult(
            data=self._validate_data(statement.entity, data, validation_mode),
            next_cursor=next_cursor,
            has_next_page=next_cursor is not None,
        )

    async def query_async(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> PaginatedResult[TViewInstance]:
        return await run_async(self.query, statement, validation_mode)

    def query_all_pages(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> list[TViewInstance]:
        if statement.get_values().cursor:
            raise ValueError("Cursor should be none when querying all pages")

        data, _ = self._cognite_adapter.query(statement, True)

        return self._validate_data(statement.entity, data, validation_mode)

    async def query_all_pages_async(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> list[TViewInstance]:
        return await run_async(self.query_all_pages, statement, validation_mode)

    def aggregate(
        self, statement: AggregationStatement[TAggregatedViewInstance]
    ) -> list[TAggregatedViewInstance]:
        data = self._cognite_adapter.aggregate(statement)

        return [statement.entity.model_validate(item) for item in data]

    async def aggregate_async(
        self, statement: AggregationStatement[TAggregatedViewInstance]
    ) -> list[TAggregatedViewInstance]:
        return await run_async(self.aggregate, statement)

    def upsert(
        self,
        entries: list[TWritableViewInstance],
        replace: bool = False,
        remove_unset: bool = False,
    ) -> None:
        if not entries:
            return

        return self._cognite_adapter.upsert(entries, replace, remove_unset)

    async def upsert_async(
        self,
        entries: list[TWritableViewInstance],
        replace: bool = False,
        remove_unset: bool = False,
    ) -> None:
        return await run_async(self.upsert, entries, replace, remove_unset)

    def delete(self, nodes: list[TViewInstance]) -> None:
        self._cognite_adapter.delete(
            nodes,
        )

    async def delete_async(self, nodes: list[TViewInstance]) -> None:
        return await run_async(self.delete, nodes)

    @classmethod
    def from_config_file(cls, config_file: str | Path) -> "Engine":
        client, dm_id = generate_engine_params(config_file)
        return Engine(client, dm_id)

    def _validate_data(
        self,
        entity: type[TViewInstance],
        data: list[dict[str, Any]],
        validation_mode: ValidationMode,
    ) -> list[TViewInstance]:
        result: list[TViewInstance] = []
        for item in data:
            try:
                validated_item = entity.model_validate(item)
                include_edges(item, validated_item)
                result.append(validated_item)
            except Exception:
                if validation_mode == "ignoreOnError":
                    continue
                raise
        return result



from typing import (
    Any,
    dataclass_transform,
)

from pydantic import BaseModel, ConfigDict, Field
from pydantic._internal import _model_construction
from pydantic.alias_generators import to_camel

from industrial_model.statements import Column


@dataclass_transform(kw_only_default=True, field_specifiers=(Field,))
class DBModelMetaclass(_model_construction.ModelMetaclass):
    is_constructing: bool = False

    def __new__(
        mcs,
        cls_name: str,
        bases: tuple[type[Any], ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> type:
        mcs.is_constructing = True
        cls = super().__new__(mcs, cls_name, bases, namespace, **kwargs)
        mcs.is_constructing = False
        return cls

    def __getattr__(self, key: str) -> Any:
        if self.is_constructing:
            return super().__getattr__(key)  # type: ignore

        try:
            return super().__getattr__(key)  # type: ignore
        except AttributeError:
            if key in self.model_fields:
                return Column(self.model_fields[key].alias or key)
            raise


class RootModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    def get_field_name(self, field_name_or_alias: str) -> str | None:
        entry = self.__class__.model_fields.get(field_name_or_alias)
        if entry:
            return field_name_or_alias

        for key, field_info in self.__class__.model_fields.items():
            if field_info.alias == field_name_or_alias:
                return key

        return None

    def get_field_alias(self, field_name_or_alias: str) -> str | None:
        entry = self.__class__.model_fields.get(field_name_or_alias)
        if entry:
            return entry.alias

        for _, field_info in self.__class__.model_fields.items():
            if field_info.alias == field_name_or_alias:
                return field_info.alias

        return None



from dataclasses import dataclass

import cognite.client.data_classes.filters as filters
from cognite.client.data_classes.aggregations import (
    Avg,
    Count,
    Max,
    MetricAggregation,
    Min,
    Sum,
)
from cognite.client.data_classes.data_modeling import (
    View,
)

from industrial_model.models import TAggregatedViewInstance
from industrial_model.statements import AggregationStatement

from .filter_mapper import (
    FilterMapper,
)
from .view_mapper import ViewMapper


@dataclass
class AggregationQuery:
    view: View
    metric_aggregation: MetricAggregation
    filters: filters.Filter | None
    group_by_columns: list[str]
    limit: int


class AggregationMapper:
    def __init__(self, view_mapper: ViewMapper):
        self._view_mapper = view_mapper
        self._filter_mapper = FilterMapper(view_mapper)

    def map(
        self, statement: AggregationStatement[TAggregatedViewInstance]
    ) -> AggregationQuery:
        root_node = statement.entity.get_view_external_id()

        root_view = self._view_mapper.get_view(root_node)

        statement_values = statement.get_values()
        filters_ = (
            filters.And(
                *self._filter_mapper.map(statement_values.where_clauses, root_view)
            )
            if statement_values.where_clauses
            else None
        )
        aggregation_property = statement_values.aggregation_property.property
        metric_aggregation: MetricAggregation | None = None
        if statement.aggregate == "avg":
            metric_aggregation = Avg(aggregation_property)
        elif statement.aggregate == "min":
            metric_aggregation = Min(aggregation_property)
        elif statement.aggregate == "max":
            metric_aggregation = Max(aggregation_property)
        elif statement.aggregate == "sum":
            metric_aggregation = Sum(aggregation_property)
        elif statement.aggregate == "count":
            metric_aggregation = Count(aggregation_property)

        if metric_aggregation is None:
            raise ValueError(f"Unsupported aggregate function: {statement.aggregate}")

        group_by_columns = (
            [prop.property for prop in statement_values.group_by_properties]
            if statement_values.group_by_properties is not None
            else statement.entity.get_group_by_fields()
        )
        return AggregationQuery(
            view=root_view,
            metric_aggregation=metric_aggregation,
            filters=filters_,
            group_by_columns=group_by_columns,
            limit=statement_values.limit,
        )



from .config import DataModelId
from .engines import AsyncEngine, Engine
from .models import (
    AggregatedViewInstance,
    InstanceId,
    PaginatedResult,
    RootModel,
    TAggregatedViewInstance,
    TViewInstance,
    TWritableViewInstance,
    ValidationMode,
    ViewInstance,
    ViewInstanceConfig,
    WritableViewInstance,
)
from .statements import aggregate, and_, col, not_, or_, search, select

__all__ = [
    "aggregate",
    "AggregatedViewInstance",
    "and_",
    "or_",
    "col",
    "not_",
    "select",
    "search",
    "ViewInstance",
    "InstanceId",
    "TViewInstance",
    "DataModelId",
    "TAggregatedViewInstance",
    "TWritableViewInstance",
    "ValidationMode",
    "Engine",
    "AsyncEngine",
    "PaginatedResult",
    "RootModel",
    "ViewInstanceConfig",
    "WritableViewInstance",
]
